/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.analysis;

import java.net.MalformedURLException;
import java.util.NoSuchElementException;
import java.util.Vector;

import quicktime.QTException;
import quicktime.std.movies.Movie;
import quicktime.std.movies.Track;
import quicktime.std.movies.media.AudioMediaHandler;
import quicktime.std.movies.media.Media;
import quicktime.std.movies.media.MediaHandler;

import net.jasonfreeman.itsm.analysis.emd.InvalidDummyCostException;
import net.jasonfreeman.itsm.analysis.emd.InvalidMatrixException;
import net.jasonfreeman.itsm.analysis.emd.InvalidWeightsVectorException;
import net.jasonfreeman.itsm.analysis.emd.NotEnoughBasicVariablesFoundException;
import net.jasonfreeman.itsm.analysis.fft.ComplexDoubleFFT;
import net.jasonfreeman.itsm.analysis.fft.ComplexDoubleFFT_Radix2;
import net.jasonfreeman.itsm.analysis.fft.FFTHelper;
import net.jasonfreeman.itsm.analysis.genetic.Genetic;
import net.jasonfreeman.itsm.gui.ProgressManager;
import net.jasonfreeman.itsm.selection.Song;
import net.jasonfreeman.itsm.util.Constants;
import net.jasonfreeman.itsm.util.ErrorHandler;
import net.jasonfreeman.itsm.util.QTSessionManager;
import net.jasonfreeman.itsm.util.SettingsManager;

/**
 * 
 * Main class for doing audio analysis of files.
 * 
 * @author jason
 *
 */

public class Analyzer {

	SettingsManager settings = SettingsManager.getInstance();

	ProgressManager progress = ProgressManager.getInstance();
	
	ComplexDoubleFFT fftAnalyzer = new ComplexDoubleFFT_Radix2(settings.getFftFrameSize());

	public Analyzer() {

	}

	/**
	 * Do the analysis.
	 * 
	 * @param songs the songs to analyze
	 * @return one chunk from each song to include in the signature.
	 */
	
	public Chunk[] doAnalysis(Song[] songs) {
		Vector chunks = new Vector();
		int length = songs.length;
		Song song = null;
		for (int i = 0; i < length; i++) {
			try {
				song = songs[i];
				progress.setTaskInfo("Analyzing " + song.getName() + " (file " + (i + 1) + " of "
						+ length + ")");
				progress.setPercentInfo(0);
				progress.setTotalInfo(2 + (i * 96 / length));
				if (!settings.isProcessing()) {
					return null;
				}
				float rank = i / (float) length;
				// open/close QT for each song analysis to avoid crashing problems in Safari
				QTSessionManager.getInstance();
				if (hasAudioTrack(songs[i])) {
					Decoder downSampler = new Decoder(songs[i], rank);
					downSampler.setSegmentSizeInSeconds(settings
							.getWindowSizeInSeconds(rank));
					selectChunk(song, downSampler, chunks, rank, downSampler.getOrigTimeBase());
					downSampler.cleanup();
				}
				QTSessionManager.shutdown();
			} catch (QTException e) {
				ErrorHandler.getInstance().handleError(e, "There was an error processing " + song.getLocation() + " for analysis.");
			} catch (MalformedURLException e) {
				ErrorHandler.getInstance().handleError(e, "There was an error processing " + song.getLocation() + " for analysis.");
			}
		}
		progress.setTaskInfo("Optimizing song chunk order");
		progress.setPercentInfo(0);
		if (!settings.isProcessing()) {
			return null;
		}
		if (chunks.size() == 0) {
			ErrorHandler.getInstance().handleWarning("iTSM was unable to open any of the files in your iTunes database! It could not create a signature.");
			return null;
		}
		return optimizeChunkOrder(chunks);
	}

	/**
	 * Perform traveling salesman optimization of chunk order.
	 * 
	 * @param chunks chunks to be reordered
	 * @return chunks in new optimized order
	 */
	private Chunk[] optimizeChunkOrder(Vector chunks) {
		Chunk[] chunkArray = new Chunk[chunks.size()];
		chunks.copyInto(chunkArray);
		if (!settings.isDoAudioAnalysis()) {
			return chunkArray; // skip optimization
		}
		Genetic travSalesman = new Genetic();
		return (rotate(travSalesman.optimizeChunks(chunkArray)));
	}

	/**
	 * Put longest-length chunk at the end so it sounds nice and conclusive...
	 * @param chunks
	 * @return
	 */
	private Chunk[] rotate(Chunk[] chunks) {
		int moveToEnd = 0;
		for (int i=1; i < chunks.length; i++) {
			if ((chunks[i].getDur() / (float) chunks[i].getTimeBase()) >
				(chunks[moveToEnd].getDur() / (float) chunks[moveToEnd].getTimeBase()))
				moveToEnd = i;
		}
		Chunk[] result = new Chunk[chunks.length];
		int index = 0;
		for (int i=moveToEnd + 1; i < chunks.length; i++) {
			result[index] = chunks[i];
			index++;
		}
		for (int i=0; i < moveToEnd + 1; i++) {
			result[index] = chunks[i];
			index++;
		}
		return result;
	}

	/**
	 * Select a single chunk from a single song such that it's most spectrally similar
	 * to the last chunk we chose.
	 * 
	 * @param song
	 * @param downSampler Decoder class for this song
	 * @param chunks all the chunks chosen so far
	 * @param rank rank of song in selection process
	 * @param timeBase time scale of this song
	 * @throws QTException
	 */
	private void selectChunk(Song song, Decoder downSampler, Vector chunks, float rank, int timeBase)
			throws QTException {
		if (!settings.isDoAudioAnalysis()) {
			selectRandomChunk(song, chunks, rank, timeBase);
			return;
		}
		if (chunks.size() == 0) {
			selectFirstChunk(song, downSampler, chunks, rank, timeBase);
			return;
		}
		int frameSize = settings.getFftFrameSize();
		int frameGroupSize = settings.getFftFrameGroupSize(rank);
		double[][] fft = new double[frameGroupSize][frameSize];
		double[] fftAvg, fftDAvg = null;
		double[] bestFitFftAvg = null;
		double[] bestFitFftDAvg = null;
		int bestFitPosition = 0;
		double[] lastFftAvg = getLastFFtAvg(chunks);
		double[] lastFftDAvg = getLastFFtDAvg(chunks);
		double diff, fftMag;
		double bestDiff = Double.MAX_VALUE;
		int bestMaxAmp = 128;
		byte[] segment;
		int i = 0;

		while ((segment = downSampler.getNextSegment()) != null) {
			if (!settings.isProcessing()) {
				return;
			}
			if ((i % 10) == 0)
				progress.setPercentInfo(downSampler.getPercent());
			for (int j = 0; j < frameGroupSize; j++) {
				doFFt(segment, fft, frameSize, j);
			}
			fftAvg = FFTHelper.doFFTAverage(fft);
			fftMag = FFTHelper.doFFTAvgMag(fftAvg);
			if (!settings.isFastMode())
				fftDAvg = FFTHelper.doFFTDAverage(fft);
			if (fftMag > settings.getMinFftMag()) {
				try {
					if (settings.isFastMode())
						diff = FFTHelper.doFFTdiff(fftAvg, lastFftAvg);
					else
						diff = FFTHelper.doFFTdiffWithD(fftAvg, lastFftAvg, fftDAvg, lastFftDAvg);
					if (diff < bestDiff) {
						bestDiff = diff;
						bestFitPosition = downSampler.getSegmentPositionInTimeCode();
						bestFitFftAvg = fftAvg;
						bestFitFftDAvg = fftDAvg;
						bestMaxAmp = findMaxAmp(segment);
					}
				} catch (InvalidMatrixException e) {
					ErrorHandler.getInstance().handleError(e);
				} catch (InvalidDummyCostException e) {
					ErrorHandler.getInstance().handleError(e);
				} catch (InvalidWeightsVectorException e) {
					ErrorHandler.getInstance().handleError(e);
				} catch (NotEnoughBasicVariablesFoundException e) {
					ErrorHandler.getInstance().handleError(e);
				} catch (NullPointerException e) {
					ErrorHandler.getInstance().handleError(e);
				}
				i++;
			}
		}
		if (bestFitFftAvg != null)
			createChunk(chunks, song, bestFitFftAvg, bestFitFftDAvg, bestFitPosition, downSampler.getSegmentSizeInTimeCode(), bestMaxAmp, timeBase);
	}

	/**
	 * Instead of doing audio analysis to determine best chunk to use, pick one at random.
	 * 
	 * @param song
	 * @param chunks
	 * @param rank
	 * @param timeBase
	 */
	private void selectRandomChunk(Song song, Vector chunks, float rank, int timeBase) {
		Movie mov;
		try {
			mov = song.openAsMovie();
			int dur = mov.getDuration();
			int segmentDur = settings.getFftFrameGroupSize(rank) * settings.getFftFrameSize()
							/ Constants.ANALYSIS_SAMPLING_RATE * timeBase;
			int startFrom = (int) (Math.random() * (dur - segmentDur));
			createChunk(chunks, song, null, null,startFrom, segmentDur, 1, timeBase);
		} catch (MalformedURLException e) {

		} catch (QTException e) {

		}
		
	}

	/**
	 * Select a chunk from the first song based on maximum spectral energy.
	 * (We need something to start from...)
	 * 
	 * @param song
	 * @param downSampler
	 * @param chunks
	 * @param rank
	 * @param timeBase
	 * @throws QTException
	 */
	private void selectFirstChunk(Song song, Decoder downSampler,
			Vector chunks, float rank, int timeBase) throws QTException {
		int frameSize = settings.getFftFrameSize();
		int frameGroupSize = settings.getFftFrameGroupSize(rank);
		double[][] fft = new double[frameGroupSize][frameSize];
		double[] fftAvg;
		double[] bestMagFftAvg = null;
		double[] bestMagFftDAvg = null;
		int bestMagPosition = 0;
		double mag;
		double bestMag = 0;
		byte[] segment;
		int bestMaxAmp = 128;
		int i = 0;

		while ((segment = downSampler.getNextSegment()) != null) {
			if (!settings.isProcessing()) {
				return;
			}
			if ((i % 10) == 0)
				progress.setPercentInfo(downSampler.getPercent());
			for (int j = 0; j < frameGroupSize; j++) {
				doFFt(segment, fft, frameSize, j);
			}
			fftAvg = FFTHelper.doFFTAverage(fft);
			
			mag = FFTHelper.doFFTAvgMag(fftAvg);
			if (mag > bestMag) {
				bestMag = mag;
				bestMagPosition = downSampler.getSegmentPositionInTimeCode();
				bestMagFftAvg = fftAvg;
				if (!settings.isFastMode())
					bestMagFftDAvg = FFTHelper.doFFTDAverage(fft);
				bestMaxAmp = findMaxAmp(segment);
			}
			i++;
		}
		if (bestMagFftAvg != null)
			createChunk(chunks, song, bestMagFftAvg, bestMagFftDAvg, bestMagPosition, downSampler.getSegmentSizeInTimeCode(), bestMaxAmp, timeBase);
	}

	/**
	 * @param chunks
	 * @param song
	 * @param fftAvg
	 * @param i
	 */
	private void createChunk(Vector chunks, Song song, double[] fftAvg, double[] fftDAvg,
			int movieTime, int chunkDur, int maxAmp, int timeBase) {
		chunks.addElement(new Chunk(song, movieTime, chunkDur, fftAvg, fftDAvg, maxAmp, timeBase));
	}

	/**
	 * @param frame
	 * @param fft
	 * @param frameSize
	 * @param j
	 */
	private void doFFt(byte[] frame, double[][] fft, int frameSize, int j) {
		int fftLength = frameSize * 2;
		double[] data = new double[fftLength];
		for (int i = 0; i < frameSize; i++) {
			data[2 * i] = frame[i + frameSize * j] / 128.0;
			data[2 * i + 1] = 0;
		}
		fft[j] = FFTHelper.doFFTAnalysis(data, frameSize);
	}

	private double[] getLastFFtAvg(Vector chunks) {
		try {
			return ((Chunk) chunks.lastElement()).getFftAvg();
		} catch (NoSuchElementException e) {
			return null;
		}
	}
	
	private double[] getLastFFtDAvg(Vector chunks) {
		try {
			return ((Chunk) chunks.lastElement()).getFftDAvg();
		} catch (NoSuchElementException e) {
			return null;
		}
	}

	private int findMaxAmp(byte[] segment) {
		int max = 0;
		int length = segment.length;
		for (int i=0; i < length; i++) {
			if (Math.abs(segment[i]) > max)
				max = Math.abs(segment[i]);
		}
		return max;
	}
	
	private boolean hasAudioTrack(Song song) {
        try {
			Movie mov = song.openAsMovie();
			for (int j = 1; j <= mov.getTrackCount(); j++) {
			    try {
					Track t = mov.getTrack(j);
					Media m = t.getMedia();
					MediaHandler mh = m.getHandler();
					if (mh instanceof AudioMediaHandler) {
						return true;
					}
				} catch (QTException e) {
					// ignore, this just means that's not a track we want...
				}
			}
			return false;
		} catch (Throwable e) {
			return false;
		}
	}
	
}