/*
 * Created on Jun 14, 2005
 *
 */
package net.jasonfreeman.itsm.analysis.fft;

import net.jasonfreeman.itsm.analysis.emd.EMD;
import net.jasonfreeman.itsm.analysis.emd.InvalidDummyCostException;
import net.jasonfreeman.itsm.analysis.emd.InvalidMatrixException;
import net.jasonfreeman.itsm.analysis.emd.InvalidWeightsVectorException;
import net.jasonfreeman.itsm.analysis.emd.NotEnoughBasicVariablesFoundException;
import net.jasonfreeman.itsm.analysis.emd.RusselInitializer;
import net.jasonfreeman.itsm.analysis.emd.Signatures;
import net.jasonfreeman.itsm.util.SettingsManager;

/**
 * 
 * FFT algorithm and utilities.
 * 
 * @author jason
 *  
 */
public class FFTHelper {

	private static EMD emd = null;
	private static int length;
	private static int[] frame1Int, frame2Int;
	private static Signatures sig;
	
	private static ComplexDoubleFFT fft = new ComplexDoubleFFT_Radix2(SettingsManager.getInstance().getFftFrameSize());
	
    /**
     * 
     * @param data data to analyze
     * @param nn 2*nn is the length of array
     * @return array of length nn which gives the magnitudes
     */
    public static double[] doFFTAnalysis(double[] data, int nn) {
        fft.transform(data);
        double[] mags = new double[nn];
        for (int i=0; i < nn; i++) {
            mags[i] = mag(data[2*i], data[2*i+1]);
        }
        return mags;
    }
    
    /**
     * Average several FFT frames together
     * 
     * @param frames the frames to average
     * @return average frame data
     */
    public static double[] doFFTAverage(double[][] frames) {
       int frameLength = frames[0].length;
       int numFrames = frames.length;
       double[] sums = new double[frameLength];
       for (int i=0; i < numFrames; i++) {
           for (int j=0; j < frameLength; j++) {
               sums[j] += frames[i][j];
           }
       }
       for (int i=0; i < frameLength; i++) {
           sums[i] = sums[i] / numFrames;
       }
       return sums;
    }

    /**
     * average derivative from one FFT frame to the next
     * 
     * @param frames
     * @return
     */
    public static double[] doFFTDAverage(double[][] frames) {
        int frameLength = frames[0].length;
        int numFrames = frames.length;
        double[] sums = new double[frameLength];
        for (int i=1; i < numFrames; i++) {
            for (int j=0; j < frameLength; j++) {
                sums[j] += frames[i][j] - frames[i-1][j];
            }
        }
        for (int i=0; i < frameLength; i++) {
        		// for EMD everything has to be positive, but we don't want to just do abs,
        		// so we recenter around 10
            sums[i] = sums[i] / numFrames + 10;
            if (sums[i] < 0)
            		sums[i] = 0;
        }
        return sums;
     }
    
    /**
     * 
     * @param data
     * @return average of all values in data
     */
    public static double doFFTAvgMag(double[] data) {
        int length = data.length;
        double sum = 0;
        for (int i=0; i < length; i++) {
            sum += Math.abs(data[i]);
        }
        return (sum / length);
    }
    
    /**
     * Compare two averaged FFT frames and derivative frames which describe how the original data was changing over time.
     * This function is not used in fastMode.
     * 
     * @param frame1
     * @param frame2
     * @param frame1D
     * @param frame2D
     * @return
     * @throws InvalidMatrixException
     * @throws InvalidDummyCostException
     * @throws InvalidWeightsVectorException
     * @throws NotEnoughBasicVariablesFoundException
     */
    public static double doFFTdiffWithD(double[] frame1, double[] frame2, double[] frame1D, double[] frame2D) throws InvalidMatrixException, InvalidDummyCostException, InvalidWeightsVectorException, NotEnoughBasicVariablesFoundException {
    		return FFTHelper.doFFTdiff(frame1, frame2) + FFTHelper.doFFTdiff(frame1D, frame2D);
    }
    
    /**
     * Use EMD algorithm to calculate similarity of frame1 and frame2.
     * Used in fastmode.
     * 
     * @param frame1 first frame to compare
     * @param frame2 second frame to compare (must be same size as frame1, this is not checked!)
     * @return similarity measure
     * @throws InvalidDummyCostException
     * @throws InvalidMatrixException
     * @throws InvalidWeightsVectorException
     * @throws NotEnoughBasicVariablesFoundException
     */
    public static double doFFTdiff(double[] frame1, double[] frame2) throws InvalidMatrixException, InvalidDummyCostException, InvalidWeightsVectorException, NotEnoughBasicVariablesFoundException {
        if (emd == null) {
        		initEMD(frame1);
        }
    		for (int i=0; i < length; i++) {
    		    frame1Int[i] = (int) (frame1[i] * 100);
    		    frame2Int[i] = (int) (frame2[i] * 100);
    		}
    		scaleFrames(frame1Int, frame2Int);
    		sig.setQWeights(frame1Int);
    		sig.setPWeights(frame2Int);
    		double e = emd.calculate(sig, (byte)1);
    		return e;
    }
    
    private static void initEMD(double[] frame) throws InvalidMatrixException, InvalidDummyCostException {
   		emd = new EMD(new RusselInitializer(), EMD.DL_NONE);
		length = frame.length;
		float[][] cost = new float[length][length];
		for (int i=0; i < length; i++) {
		    for (int j=0; j < length; j++) {
		        cost[i][j] = Math.abs(i - j);
		    }
		}
		frame1Int = new int[length];
		frame2Int = new int[length];
		sig = new Signatures(cost);
    }
    
    /**
     * EMD requires the sum of all elements in both arrays be the same!
     * So we do some scaling here to make sure that's true.
     * 
     * @param frame1Int
     * @param frame2Int
     */
    private static void scaleFrames(int[] frame1Int, int[] frame2Int) {
        int sum1 = 0;
        int sum2 = 0;
        for (int i=0; i < length; i++) {
            sum1 += frame1Int[i];
            sum2 += frame2Int[i];
        }
        float factor = sum1 / (float) sum2;
        for (int i=0; i < length; i++) {
            frame2Int[i] = (int) (frame2Int[i] * factor);
        }
    }

    private static double mag(double real, double imag) {
        if ((real < 0.0000001) && (imag < 0.0000001))
            return 0;
        // we should return Math.sqrt of this,
        // but to save CPU, we don't, since it won't
        // affect any other calculations we make
        return real * real + imag * imag;
    }
    
    
}