/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * This source file adapted from:
 * http://pcbunn.cacr.caltech.edu/Java/Genetic.html
 * 
 */

package net.jasonfreeman.itsm.analysis.genetic;

import net.jasonfreeman.itsm.analysis.Chunk;

public class Chromosome {
	int [] chunk_list;
	double cost;
	double Mutation_probability;
	int length;
	int cut_length;

	Chromosome(Chunk[] chunks, CachingCostMatrix costMatrix) {
		length = chunks.length;
		boolean taken[] = new boolean[length];
		chunk_list = new int[length];
		cost = 0.0;
		for(int i=0;i<length;i++) taken[i] = false;
		for(int i=0;i<length-1;i++) {
			int icandidate;
			do {
				icandidate = (int) ( 0.999999* Math.random() * (double) length );
			} while(taken[icandidate]);
			chunk_list[i] = icandidate;
			taken[icandidate] = true;
			if(i == length-2) {
				icandidate = 0;
				while(taken[icandidate]) icandidate++;
				chunk_list[i+1] = icandidate;
			}
		}
		calculate_cost(chunks, costMatrix);
		cut_length = 1;
	}

	void calculate_cost(Chunk[] chunks, CachingCostMatrix costMatrix) {
		cost = 0;
		for(int i=0;i<length-1;i++) {
			cost += costMatrix.getCost(i, i+1);
		}
	}


	double get_cost() {
		return cost;
	}

	int get_city(int i) {
		return chunk_list[i];
	}

	void set_cities(int [] list) {
		for(int i=0;i<length;i++) {
			chunk_list[i] = list[i];
		}
	}

	void set_city(int index, int value) {
		chunk_list[index] = value;
	}

	void set_cut(int cut) {
		cut_length = cut;
	}

	void set_mutation(double prob) {
		Mutation_probability = prob;
	}

	void print() {
		System.out.print("chromosome: ");
		for(int i=0;i<length;i++) System.out.print(" "+chunk_list[i]);
		System.out.println(" ");
	}

	boolean check() {
		boolean taken[] = new boolean[length];
		for(int i=0;i<length;i++) taken[i] = false;
		for(int i=0;i<length;i++) taken[chunk_list[i]] = true;
		for(int i=0;i<length;i++) {
			if(!taken[i]) {
				System.out.print("Bad !"); print();
				return false;
			}
		}
		return true;
	}

	int mate(Chromosome father, Chromosome offspring1, Chromosome offspring2) {
		int cutpoint1 = (int) (0.999999*Math.random()*(double)(length-cut_length));
		int cutpoint2 = cutpoint1 + cut_length;

		//System.out.print("\nMother: "); print();
		//System.out.print("\nFather: "); father.print();
		//System.out.println("\nCutpoints: "+cutpoint1+" "+cutpoint2);

		boolean taken1 [] = new boolean[length];
		boolean taken2 [] = new boolean[length];
		int off1 [] = new int[length];
		int off2 [] = new int[length];

		for(int i=0;i<length;i++) {
			taken1[i] = false;
			taken2[i] = false;
		}

		for(int i=0;i<length;i++) {
			if(i<cutpoint1 || i>= cutpoint2) {
				off1[i] = -1;
				off2[i] = -1;
			} else {
				int imother = chunk_list[i];
				int ifather = father.get_city(i);
				off1[i] = ifather;
				off2[i] = imother;
				taken1[ifather] = true;
				taken2[imother] = true;
			}
		}

			
		//System.out.print("\nOff1  : "); offspring1.print();
		//System.out.print("\nOff2  : "); offspring2.print();

		for(int i=0;i<cutpoint1;i++) {
			if(off1[i] == -1) {
				for(int j=0;j<length;j++) {
					int imother = chunk_list[j];
					if(!taken1[imother]) {
						off1[i] = imother; 
						taken1[imother] = true;
						break;
					}
				}
			}
			if(off2[i] == -1) {
				for(int j=0;j<length;j++) {
					int ifather = father.get_city(j);
					if(!taken2[ifather]) {
						off2[i] = ifather;
						taken2[ifather] = true;
						break;
					}
				}
			}
		}
		for(int i=length-1;i>=cutpoint2;i--) {
			if(off1[i] == -1) {
				for(int j=length-1;j>=0;j--) {
					int imother = chunk_list[j];
					if(!taken1[imother]) {
						off1[i] = imother; 
						taken1[imother] = true;
						break;
					}
				}
			}
			if(off2[i] == -1) {
				for(int j=length-1;j>=0;j--) {
					int ifather = father.get_city(j);
					if(!taken2[ifather]) {
						off2[i] = ifather;
						taken2[ifather] = true;
						break;
					}
				}
			}
		}




		offspring1.set_cities(off1);
		offspring2.set_cities(off2);


		//System.out.print("\nOff1  : "); offspring1.print();
		//System.out.print("\nOff2  : "); offspring2.print();



		int mutate = 0;
		if(Math.random() < Mutation_probability) {
				int iswap1 = (int) (0.999999*Math.random()*(double)(length));
				int iswap2 = (int) (0.999999*Math.random()*(double)length);
				int i = off1[iswap1];
				off1[iswap1] = off1[iswap2];
				off1[iswap2] = i;
			mutate++;
		}
		if(Math.random() < Mutation_probability) {
				int iswap1 = (int) (0.999999*Math.random()*(double)(length));
				int iswap2 = (int) (0.999999*Math.random()*(double)length);
				int i = off2[iswap1];
				off2[iswap1] = off2[iswap2];
				off2[iswap2] = i;
			mutate++;
		}
		//System.out.print("\nOff1  : "); offspring1.print();
		//System.out.print("\nOff2  : "); offspring2.print();
		return mutate;
	}



	int oldmate(Chromosome father, Chromosome offspring1, Chromosome offspring2) {
		int cutpoint1 = (int) (0.999999*Math.random()*(double)(length-cut_length));
		int cutpoint2 = cutpoint1 + cut_length;
		//System.out.println("Cutpoints: "+cutpoint1+" "+cutpoint2);
		boolean taken1 [] = new boolean[length];
		boolean taken2 [] = new boolean[length];
		for(int i=0;i<length;i++) {
			taken1[i] = false;
			taken2[i] = false;
		}
		for(int i=0;i<length;i++) {
			int imother = chunk_list[i];
			int ifather = father.get_city(i);
			int icand = imother;
			if( i>=cutpoint1 && i<cutpoint2) icand = ifather;
			if(taken1[icand]) {
				icand = 0;
				while(taken1[icand]) icand++;
			}
			offspring1.set_city(i,icand);
			taken1[icand] = true;

			icand = ifather;
			if( i>=cutpoint1 && i<cutpoint2) icand = imother;
			if(taken2[icand]) {
				icand = 0;
				while(taken2[icand]) icand++;
			}
			offspring2.set_city(i,icand);
			taken2[icand] = true;
		}
		int mutate = 0;
		if(Math.random() < Mutation_probability) {
			int iswap1 = (int) (0.999999*Math.random()*(double)length);
			int iswap2 = (int) (0.999999*Math.random()*(double)length);
			int i = offspring1.get_city(iswap1);
			offspring1.set_city(iswap1,offspring1.get_city(iswap2));
			offspring1.set_city(iswap2,i);
			mutate++;
		}
		if(Math.random() < Mutation_probability) {
			int iswap1 = (int) (0.999999*Math.random()*(double)length);
			int iswap2 = (int) (0.999999*Math.random()*(double)length);
			int i = offspring2.get_city(iswap1);
			offspring2.set_city(iswap1,offspring2.get_city(iswap2));
			offspring2.set_city(iswap2,i);
			mutate++;
		}
		//System.out.print("\nMother: "); print();
		//System.out.print("\nFather: "); father.print();
		//System.out.print("\nOff1  : "); offspring1.print();
		//System.out.print("\nOff2  : "); offspring2.print();
		return mutate;
	}
}
