/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.selection;

import java.io.File;
import java.io.IOException;
import java.io.StringBufferInputStream;
import java.util.LinkedList;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

import net.jasonfreeman.itsm.gui.ProgressManager;
import net.jasonfreeman.itsm.util.ErrorHandler;
import net.jasonfreeman.itsm.util.SettingsManager;

/**
 * Identify and select and ordered list of files to process. This class does not
 * use DOM but instead uses SAX directly to avoid having to read the whole XML
 * file into memory at once. Those iTunes XML files can get really large!!! But
 * that means that this code is not nearly as readable or elegant. Sorry!
 * 
 * @author jason
 * 
 */
public class FileSelector implements org.xml.sax.ErrorHandler, ContentHandler {

	private String xml;

	private int maxSongs;

	private ProgressManager progress;

	private SongComparer songComparer;

	private Song song = null;

	private Song[] winners;

	private int lastIndex = -1;

	private int elementsTraversed = 0;

	private LinkedList elementQ = new LinkedList();

	private String currentKey = null;

	private String currentVal = null;

	public FileSelector(String xml, int maxSongs) {
		this.maxSongs = maxSongs;
		this.xml = xml;
		winners = new Song[maxSongs];
		progress = ProgressManager.getInstance();
		setSongComparer();
	}

	public String getXml() {
		return xml;
	}

	public void setXml(String xml) {
		this.xml = xml;
	}

	public int getMaxSongs() {
		return maxSongs;
	}

	public void setMaxSongs(int maxSongs) {
		this.maxSongs = maxSongs;
	}

	public Song[] selectFiles() {

		try {
			SAXParserFactory spf = SAXParserFactory.newInstance();
			spf.setValidating(false);
			SAXParser saxParser = spf.newSAXParser();
			XMLReader xmlReader = saxParser.getXMLReader();
			xmlReader.setContentHandler(this);
			xmlReader.setErrorHandler(this);
			// prevent crimson from downloading plist DTD from apple.com
			xmlReader.setEntityResolver(new EntityResolver() {
				public InputSource resolveEntity(String arg0, String arg1)
						throws SAXException, IOException {
					return new InputSource(new StringBufferInputStream(""));
				}
			});
			xmlReader.parse(convertToFileURL(xml));
		} catch (FactoryConfigurationError e) {
			ErrorHandler.getInstance().handleError(e);
		} catch (ParserConfigurationException e) {
			ErrorHandler.getInstance().handleError(e);
		} catch (SAXException e) {
			ErrorHandler.getInstance().handleError(e);
		} catch (IOException e) {
			ErrorHandler.getInstance().handleError(e);
		}
		if (elementsTraversed == 0) {
			ErrorHandler
					.getInstance()
					.handleWarning(
							"iTSM could not find any music in your collection, except possibly some encrypted files purchased from the iTunes Music Store. It cannot create a signature!");
			return null;
		} else if (elementsTraversed < maxSongs) {
			if (SettingsManager.getInstance().getLimitMode() != SettingsManager.NO_LIMITS) {
				// repeat search without limits to get more songs
				SettingsManager.getInstance().setLimitMode(
						SettingsManager.NO_LIMITS);
				FileSelector fs = new FileSelector(xml, maxSongs);
				Song[] secondTryWinners = fs.selectFiles();
				if (secondTryWinners.length > elementsTraversed) {
					ErrorHandler
							.getInstance()
							.handleWarning(
									"iTSM could only find "
											+ elementsTraversed
											+ " songs in your library when limiting itself to one song per artist or album.\n\n"
											+ "You asked iTSM to find " + maxSongs + " songs.\n\n"
											+ "So iTSM included multiple songs by the same artist or album and was able to select "
											+ secondTryWinners.length + " songs.");
					return secondTryWinners;
				}
			}
			Song[] smaller = new Song[elementsTraversed];
			for (int i = 0; i < elementsTraversed; i++)
				smaller[i] = winners[i];
			winners = smaller;
		}
		progress.setPercentInfo(100);
		return winners;
	}

	/**
	 * This function gets called once we've parsed all the XML data for a single
	 * song and created a new song object to sort.
	 * 
	 * @param song
	 */
	private void handleSong(Song song) {
		if (isUsable(song)) {
			lastIndex = placeSong(winners, song, lastIndex);
			elementsTraversed++;
			if ((elementsTraversed % 100) == 0) {
				progress.setPercentInfo(elementsTraversed * 100 / 10000);
			}
		}
	}

	/**
	 * Sets which song comparer object will be used to sort.
	 */
	private void setSongComparer() {
		int selectionMode = SettingsManager.getInstance().getSelectionMode();
		if (selectionMode == SettingsManager.DATE_ADDED)
			songComparer = new DateAddedComparer();
		else if (selectionMode == SettingsManager.LAST_DATE_PLAYED)
			songComparer = new LastPlayedComparer();
		else if (selectionMode == SettingsManager.MY_RATING)
			songComparer = new RatingComparer();
		else if (selectionMode == SettingsManager.TIMES_PLAYED)
			songComparer = new PlayCountComparer();
		else {
			ErrorHandler.getInstance().handleError(
					new Error("invalid selectionMode ID, setting to default"));
			songComparer = new PlayCountComparer();
		}
	}

	/**
	 * Eliminate DRM'd and streamed songs from use.
	 * 
	 * @param song
	 *            song to check
	 * @return true if it's ok
	 */
	private boolean isUsable(Song song) {
		if (song.getKind().indexOf("Protected") != -1) {
			ProgressManager.getInstance().showProtectedWarning();
		}
		return (song.getLocation().startsWith("file:")
				&& (song.getKind().indexOf("Protected") == -1)
				&& (!SettingsManager.getInstance().isIgnoreVideos() || ((song
						.getKind().indexOf("movie") == -1) && (song.getKind()
						.indexOf("video") == -1))) && (!SettingsManager
				.getInstance().isIgnorePodcasts() || !song.getGenre().equals(
				"Podcast")));
	}

	/**
	 * Places a song in correct sort location in array.
	 * 
	 * @param winners
	 * @param song
	 * @param songsPlaced
	 * @return index where song was placed (if it's not placed, index will be
	 *         out of array bounds).
	 */
	private int placeSong(Song[] winners, Song song, int lastIndex) {
		// if album/artist matching is on, then check if matching song is
		// already in the list
		for (int i = 0; i <= lastIndex; i++) {
			if (match(song, winners[i])) {
				elementsTraversed--;
				if (songComparer.betterThan(song, winners[i]))
					return replaceSong(winners, song, i, lastIndex);
				else
					return lastIndex;
			}
		}
		// if album/artist doesn't match, then just put it in using normal sort
		for (int i = 0; i <= lastIndex; i++) {
			if (songComparer.betterThan(song, winners[i])) {
				return insertSong(winners, song, i, lastIndex);
			}
		}
		if (lastIndex < (maxSongs - 1)) {
			lastIndex++;
			winners[lastIndex] = song;
		}
		return lastIndex;
	}

	/**
	 * 
	 * @param song1
	 * @param song2
	 * @return true if songs match according to artist or album (as specified in
	 *         settingsmanager)
	 */
	private boolean match(Song song1, Song song2) {
		int limitMode = SettingsManager.getInstance().getLimitMode();
		if (limitMode == SettingsManager.NO_LIMITS)
			return false;
		else if (limitMode == SettingsManager.LIMIT_BY_ALBUM)
			return (song1.getAlbum().equalsIgnoreCase(song2.getAlbum()) && (song1
					.getAlbum().length() > 0));
		else if (limitMode == SettingsManager.LIMIT_BY_ARTIST)
			return (song1.getArtist().equalsIgnoreCase(song2.getArtist()) && (song1
					.getArtist().length() > 0));
		else if (limitMode == SettingsManager.LIMIT_BY_ALBUM_OR_ARTIST)
			return ((song1.getAlbum().equalsIgnoreCase(song2.getAlbum()) && (song1
					.getAlbum().length() > 0)) || (song1.getArtist()
					.equalsIgnoreCase(song2.getArtist()) && (song1.getArtist()
					.length() > 0)));
		else
			return false;
	}

	/**
	 * Replace song in winner array.
	 * 
	 * @param winners
	 * @param song
	 * @param i
	 * @param lastIndex
	 * @return
	 */
	private int replaceSong(Song[] winners, Song song, int i, int lastIndex) {
		winners[i] = song;
		return lastIndex;
	}

	/**
	 * Insert song into winner array.
	 * 
	 * @param winners
	 * @param song
	 * @param i
	 * @param lastIndex
	 * @return
	 */
	private int insertSong(Song[] winners, Song song, int i, int lastIndex) {
		if (lastIndex < (winners.length - 1))
			lastIndex++;
		for (int j = lastIndex; j > i; j--) {
			winners[j] = winners[j - 1];
		}
		winners[i] = song;
		return lastIndex;
	}

	/**
	 * Convert from a filename to a file URL.
	 */
	private static String convertToFileURL(String filename) {
		// On JDK 1.2 and later, simplify this to:
		// "path = file.toURL().toString()".
		String path = new File(filename).getAbsolutePath();
		if (File.separatorChar != '/') {
			path = path.replace(File.separatorChar, '/');
		}
		if (!path.startsWith("/")) {
			path = "/" + path;
		}
		return "file:" + path;
	}

	public void warning(SAXParseException arg0) throws SAXException {
		ErrorHandler.getInstance().handleError(arg0);
	}

	public void error(SAXParseException arg0) throws SAXException {
		ErrorHandler.getInstance().handleError(arg0);
	}

	public void fatalError(SAXParseException arg0) throws SAXException {
		ErrorHandler.getInstance().handleError(arg0);
	}

	public void setDocumentLocator(Locator arg0) {
		// do nothing
	}

	public void startDocument() throws SAXException {
		// do nothing
	}

	public void endDocument() throws SAXException {
		// do nothing
	}

	public void startPrefixMapping(String arg0, String arg1)
			throws SAXException {
		// do nothing
	}

	public void endPrefixMapping(String arg0) throws SAXException {
		// do nothing
	}

	public void startElement(String uri, String localName, String qName,
			Attributes att) throws SAXException {
		try {
			elementQ.addFirst(qName);
			if (startedTrack()) {
				song = new Song();
			} else if (isTrackKey() || isTrackValue()) {
				currentVal = "";
			}
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
		}
	}

	public void endElement(String uri, String localName, String qName)
			throws SAXException {
		try {
			if (isTrackKey()) {
				currentKey = currentVal;
			} else if (isTrackValue()) {
				if (currentKey.equals("Date Added")) {
					song.setDateCreated(currentVal);
				} else if (currentKey.equals("Play Count")) {
					song.setPlayCount(Integer.parseInt(currentVal));
				} else if (currentKey.equals("Rating")) {
					song.setRating(Integer.parseInt(currentVal));
				} else if (currentKey.equals("Play Date UTC")) {
					song.setLastPlayed(currentVal);
				} else if (currentKey.equals("Location")) {
					song.setLocation(currentVal);
				} else if (currentKey.equals("Name")) {
					song.setName(currentVal);
				} else if (currentKey.equals("Genre")) {
					song.setGenre(currentVal);
				} else if (currentKey.equals("Kind")) {
					song.setKind(currentVal);
				} else if (currentKey.equals("Album")) {
					song.setAlbum(currentVal);
				} else if (currentKey.equals("Artist")) {
					song.setArtist(currentVal);
				}
			}
			elementQ.removeFirst();
			if (finishedTrack()) {
				handleSong(song);
				song = null;
			}
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
		}
	}

	public void characters(char[] ch, int start, int length)
			throws SAXException {
		try {
			if (isTrackKey() || isTrackValue())
				currentVal += new String(ch, start, length);
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
		}
	}

	public void ignorableWhitespace(char[] arg0, int arg1, int arg2)
			throws SAXException {
		// do nothing
	}

	public void processingInstruction(String arg0, String arg1)
			throws SAXException {
		// do nothing
	}

	public void skippedEntity(String arg0) throws SAXException {
		// do nothing
	}

	private boolean startedTrack() {
		int depth = elementQ.size();
		if (depth != 4)
			return false;
		if (!elementQ.get(0).equals("dict"))
			return false;
		if (!elementQ.get(1).equals("dict"))
			return false;
		if (!elementQ.get(2).equals("dict"))
			return false;
		if (!elementQ.get(3).equals("plist"))
			return false;
		return true;
	}

	private boolean finishedTrack() {
		if (song == null)
			return false;
		int depth = elementQ.size();
		if (depth != 3)
			return false;
		if (!elementQ.get(0).equals("dict"))
			return false;
		if (!elementQ.get(1).equals("dict"))
			return false;
		if (!elementQ.get(2).equals("plist"))
			return false;
		return true;
	}

	private boolean isTrackKey() {
		if (song == null)
			return false;
		int depth = elementQ.size();
		if (depth != 5)
			return false;
		if (!elementQ.get(0).equals("key"))
			return false;
		if (!elementQ.get(1).equals("dict"))
			return false;
		if (!elementQ.get(2).equals("dict"))
			return false;
		if (!elementQ.get(3).equals("dict"))
			return false;
		if (!elementQ.get(4).equals("plist"))
			return false;
		return true;
	}

	private boolean isTrackValue() {
		if (song == null)
			return false;
		int depth = elementQ.size();
		if (depth != 5)
			return false;
		if ((!elementQ.get(0).equals("string"))
				&& (!elementQ.get(0).equals("integer"))
				&& (!elementQ.get(0).equals("date")))
			return false;
		if (!elementQ.get(1).equals("dict"))
			return false;
		if (!elementQ.get(2).equals("dict"))
			return false;
		if (!elementQ.get(3).equals("dict"))
			return false;
		if (!elementQ.get(4).equals("plist"))
			return false;
		return true;
	}
}
