/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Vector;

import javax.swing.JProgressBar;

import quicktime.QTException;
import quicktime.io.IOConstants;
import quicktime.io.OpenMovieFile;
import quicktime.io.QTFile;
import quicktime.io.QTIOException;
import quicktime.sound.SoundConstants;
import quicktime.std.StdQTConstants;
import quicktime.std.comp.ComponentDescription;
import quicktime.std.comp.ComponentIdentifier;
import quicktime.std.movies.Movie;
import quicktime.std.movies.media.SoundDescription;
import quicktime.std.qtcomponents.MovieExporter;
import quicktime.util.QTBuild;
import quicktime.util.QTUtils;

/**
 * Handles all HTTP communications with the server, including session and error logging
 * and file upload.
 * 
 * @author jason
 *
 */

public class RemoteDatabase {

	public static final String[] PROPERTIES = { "java.vendor", "java.version",
			"os.arch", "os.name", "os.version" };

	String authString;

	ServerConnectionThread serverThread = null;

	boolean canceled = false;

	/**
	 * Initializes a session with the HTTP server and gets a session id.
	 *
	 */
	public RemoteDatabase() {
		if (!SettingsManager.getInstance().useRemoteDb()) {
			return;
		}
		serverThread = new ServerConnectionThread();
		serverThread.start();
		ServerAction act = new ServerAction();
		act.type = ServerAction.TYPE_INIT;
		StringBuffer temp = new StringBuffer();
		int numParams = PROPERTIES.length + 2;
		act.names = new String[numParams];
		act.values = new String[numParams];
		for (int i = 0; i < PROPERTIES.length; i++) {
			act.names[i] = PROPERTIES[i];
			act.values[i] = System.getProperty(PROPERTIES[i]);
			temp.append(System.getProperty(PROPERTIES[i]));
		}
		temp.append(QTBuild.info());
		act.names[PROPERTIES.length] = "quicktime.version";
		act.values[PROPERTIES.length] = QTBuild.info();
		Md5 md5 = new Md5(SettingsManager.getInstance().getHostAuth()
				+ temp.toString() + SettingsManager.getInstance().getHostAuth());
		md5.processString();
		authString = md5.getStringDigest();
		act.names[PROPERTIES.length + 1] = "auth";
		act.values[PROPERTIES.length + 1] = authString;
		act.files = null;
		act.waitForResponse = true;
		serverThread.addAction(act);
	}

	/**
	 * Updates or adds logging or error information to an active session.
	 * 
	 * @param name
	 * @param value
	 */
	public void sendParameter(String name, String value) {
		if (!SettingsManager.getInstance().useRemoteDb()) {
			return;
		}
		ServerAction act = new ServerAction();
		act.type = ServerAction.TYPE_SEND_PARAM;
		String[] names = { "auth", "ssid", name };
		String[] values = { authString, serverThread.getSsid(), value };
		act.names = names;
		act.values = values;
		act.files = null;
		serverThread.addAction(act);
	}

	/**
	 * Uploads a file to the signature gallery.
	 * 
	 * @param name display name for file
	 * @param file actual file
	 * @param prog GUI progress indicator
	 */
	public void sendFile(String name, File file, JProgressBar prog) {
		if (!SettingsManager.getInstance().useRemoteDb()) {
			return;
		}
		if (!file.exists()) {
			ErrorHandler.getInstance().handleWarning("Sorry, but it looks like you moved or deleted your " + "" +
					"signature file from the location where we saved it. Since we can't find it, we can't upload it to the gallery.");
			return;
		}
		File compressedFile = null;
		try {
			compressedFile = compressFile(file);
		} catch (QTException e2) {
			ErrorHandler.getInstance().handleError(e2, "Could not compress file for upload.");
			return;
		} catch (IOException e) {
			ErrorHandler.getInstance().handleError(e, "Could not create temp file for upload.");
			return;
		}
		File[] files = null;
		try {
			files = splitFile(compressedFile);
		} catch (IOException e1) {
			ErrorHandler.getInstance().handleError(e1, "Could not split file for upload.");
			return;
		}
		ServerAction act = new ServerAction();
		act.type = ServerAction.TYPE_SEND_FILE;
		String[] names = { "auth", "ssid", "fileName", "fileSize" };
		String[] values = { authString, serverThread.getSsid(), name,
				String.valueOf(compressedFile.length()) };
		act.names = names;
		act.values = values;
		act.files = files;
		if (prog != null)
			act.progressBar = prog;
		serverThread.addAction(act);
		while ((act.status != ServerAction.STATUS_DONE)
				&& (act.status != ServerAction.STATUS_FAILED)
				&& (act.status != ServerAction.STATUS_ABORTED))
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
	    if (act.status != ServerAction.STATUS_DONE)
			ErrorHandler.getInstance().handleWarning("iTSM could not connect to the server to upload your signature; either because you are behind a proxy/firewall, you are not connected to the Internet, or the server is down.");
		deleteTempFiles(files, compressedFile);
	}

	/**
	 * Cancel any HTTP requests in queue or in progress.
	 *
	 */
	public void cancel() {
		canceled = true;
		serverThread.cancelEverything();
	}

	/**
	 * Returns current canceled state and resets canceled to false.
	 * @return true if HTTP requests have been canceled
	 */
	public boolean checkAndResetCanceled() {
		boolean temp = canceled;
		canceled = false;
		return temp;
	}

	/**
	 * Shuts down the server thread
	 * 
	 * @param wait block until server thread shuts down
	 */
	public void shutdown(boolean wait) {
		if (serverThread != null)
			serverThread.shutdown(wait);
	}

	
	/**
	 * Compress a file for faster upload to server.
	 * @param in file to compress
	 * @return link to temp compressed file
	 * @throws QTException
	 * @throws IOException
	 */
	public File compressFile(File in) throws QTException, IOException {
		File temp = File.createTempFile("iTSM", ".wav");
		QTFile saveFile = new QTFile(temp);
		
		/**
		 * In an ideal world, we'd export to an MP3 right here. But Quicktime doesn't
		 * export MP3 anymore, and on some versions you need to go Pro to do AAC.
		 * So we export to IMA-compressed AIFF instead (don't want a huge file to upload),
		 * and then it gets converted to MP3 when it arrives on the server.
		 */
		ComponentDescription cd =
			new ComponentDescription(
				StdQTConstants.movieExportType,
				StdQTConstants.kQTFileTypeAIFF,
				StdQTConstants.soundMediaType,
				0,
				0);
		ComponentIdentifier ci = ComponentIdentifier.find(cd);
		MovieExporter exp = new MovieExporter(ci);
		SoundDescription sd = new SoundDescription(QTUtils.toOSType("ima4"));
		sd.setNumberOfChannels(1);
		sd.setSampleRate(44100);
		sd.setSampleSize(16);
		sd.setCompressionID(SoundConstants.kIMACompression);
		sd.lock();
		exp.setSampleDescription(sd, StdQTConstants.soundMediaType);
		
		QTFile qtf = new QTFile(in);
        Movie mov = null;
        OpenMovieFile omf = OpenMovieFile.asRead(qtf);
        mov = Movie.fromFile(omf);
        omf.close();
        
		mov.convertToFile(null, saveFile, StdQTConstants.kQTFileTypeAIFF,
				StdQTConstants.kMoviePlayer, IOConstants.smSystemScript,
				StdQTConstants.createMovieFileDeleteCurFile
						| StdQTConstants.movieFileSpecValid, exp);
		sd.unlock();
		return temp;
	}
	
	/**
	 * Split file into multiple parts to avoid going over server's file size
	 * limits.
	 * 
	 * @param in
	 * @param out
	 * @return
	 * @throws IOException
	 * @throws Exception
	 */
	private File[] splitFile(File in) throws IOException {
		Vector files = new Vector();
		File temp = File.createTempFile("iTSM", ".wav");
		FileInputStream fis = new FileInputStream(in);
		FileOutputStream fos = new FileOutputStream(temp);
		byte[] buf = new byte[8092];
		int len = 0;
		long bytesRead = 0;
		while ((len = fis.read(buf)) != -1) {
			fos.write(buf, 0, len);
			bytesRead += len;
			if (bytesRead > 1048576) {
				files.addElement(temp);
				fos.close();
				temp = File.createTempFile("iTSM", ".wav");
				fos = new FileOutputStream(temp);
				bytesRead = 0;
			}
		}
		fis.close();
		if (bytesRead > 0) {
			files.addElement(temp);
			fos.close();
		}
		File[] fileArray = new File[files.size()];
		for (int i=0; i < files.size(); i++) {
			fileArray[i] = (File) files.elementAt(i);
		}
		return fileArray;
	}
	
	/**
	 * Delete temp files created when compressing and splitting files for upload.
	 * @param files
	 * @param file
	 */
	private void deleteTempFiles(File[] files, File file) {
		if (file != null)
			file.delete();
		if (files == null)
			return;
		for (int i=0; i < files.length; i++) {
			files[i].delete();
		}
	}
}
